!/*****************************************************************************/
! *
! *  Elmer/Ice, a glaciological add-on to Elmer
! *  http://elmerice.elmerfem.org
! *
! * 
! *  This program is free software; you can redistribute it and/or
! *  modify it under the terms of the GNU General Public License
! *  as published by the Free Software Foundation; either version 2
! *  of the License, or (at your option) any later version.
! * 
! *  This program is distributed in the hope that it will be useful,
! *  but WITHOUT ANY WARRANTY; without even the implied warranty of
! *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! *  GNU General Public License for more details.
! *
! *  You should have received a copy of the GNU General Public License
! *  along with this program (in file fem/GPL-2); if not, write to the 
! *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
! *  Boston, MA 02110-1301, USA.
! *
! *****************************************************************************/
! ******************************************************************************
! *
! *  Authors: Olivier Gagliardini, Ga¨el Durand
! *  Email:   
! *  Web:     http://elmerice.elmerfem.org
! *
! *  Original Date: 
! * 
! *****************************************************************************
!> Solver for creating a mask on whether the lower side of an ice sheet/shelf is
!>  grounded or not. +1=grounded,-1=detached, 0=grounding line (=last grounded node)
SUBROUTINE GroundedSolver( Model,Solver,dt,TransientSimulation )
!------------------------------------------------------------------------------
!******************************************************************************
!
!  For the bottom surface, creates and updates a mask which may be equal to -1, 0 or 1

!  GroundedMask = + 1 if grounded
!               = - 1 if floating
!               = 0   if on the grounding line (also grounded)
!
!  Consequently, a node is grounded if GroundedMask >= 0
!
!  y is the vertical in 2D ; z is the vertical in 3D
!
!  ARGUMENTS:
!
!  TYPE(Model_t) :: Model,  
!     INPUT: All model information (mesh, materials, BCs, etc...)
!
!  TYPE(Solver_t) :: Solver
!     INPUT: Linear & nonlinear equation solver options
!
!  REAL(KIND=dp) :: dt,
!     INPUT: Timestep size for time dependent simulations
!
!  LOGICAL :: TransientSimulation
!     INPUT: Steady state or transient simulation
!
!******************************************************************************
  USE DefUtils

  IMPLICIT NONE
!------------------------------------------------------------------------------
  TYPE(Solver_t) :: Solver
  TYPE(Model_t) :: Model

  REAL(KIND=dp) :: dt
  LOGICAL :: TransientSimulation

!------------------------------------------------------------------------------
! Local variables
!------------------------------------------------------------------------------

  TYPE(Element_t),POINTER :: Element
  TYPE(ValueList_t), POINTER :: Material, SolverParams
  TYPE(Variable_t), POINTER :: PointerToVariable
  TYPE(Nodes_t), SAVE :: Nodes

  LOGICAL :: AllocationsDone = .FALSE., GotIt, stat, Parallel, Quadratic

  INTEGER :: i, mn, n, t, Nn, istat, DIM, MSum, ZSum
  INTEGER, POINTER :: Permutation(:)

  REAL(KIND=dp), POINTER :: VariableValues(:)
  REAL(KIND=dp) :: z, toler
  REAL(KIND=dp), ALLOCATABLE :: zb(:)

  CHARACTER(LEN=MAX_NAME_LEN) :: SolverName = 'GroundedSolver'
       
  SAVE AllocationsDone, DIM, SolverName, zb
  SAVE toler, Parallel, Quadratic
  !------------------------------------------------------------------------------

  PointerToVariable => Solver % Variable
  Permutation  => PointerToVariable % Perm
  VariableValues => PointerToVariable % Values

  CALL INFO(SolverName, 'Computing grounded mask from geometry', level=3)

  !--------------------------------------------------------------
  !Allocate some permanent storage, this is done first time only:
  !--------------------------------------------------------------

  IF ( (.NOT. AllocationsDone) .OR. Solver % Mesh % Changed ) THEN
    DIM = CoordinateSystemDimension()
    mn = Solver % Mesh % MaxElementNodes

    IF (AllocationsDone) DEALLOCATE(zb)

    ALLOCATE(zb(mn), STAT=istat )
          
    IF ( istat /= 0 ) THEN
      CALL FATAL( SolverName, 'Memory allocation error.' )
    END IF

    AllocationsDone = .TRUE.
    CALL INFO( SolverName, 'Memory allocation done.',Level=1 )

    !---------------------------
    ! Environnement parallele ?
    !---------------------------
    Parallel = .FALSE.
    IF ( ParEnv % PEs > 1 ) Parallel = .TRUE.

    ! quadratic or linear elements (not used for now)
    Quadratic = .FALSE.
    Element => GetActiveElement(1)
    n = GetElementNOFNodes()
    IF ( Element % Type % ElementCode/n == 102 ) Quadratic = .TRUE.

  END IF
  ! End first time

  !--------------------------------------------
  ! First loop, where y > zb GroundedMask = -1 
  !--------------------------------------------
  DO t = 1, Solver % NumberOfActiveElements
    Element => GetActiveElement(t)
!     IF (ParEnv % myPe .NE. Element % partIndex) CYCLE
    n = GetElementNOFNodes()

    SolverParams => GetSolverParams()
    toler = GetConstReal(SolverParams, 'Toler', GotIt)
    IF (.NOT.GotIt) THEN
      CALL FATAL(SolverName, 'No tolerance given for the Grounded Mask.')
    END IF

    Material => GetMaterial( Element )
    zb(1:n) = ListGetReal( Material,'Min Zs Bottom',n , & 
                   Element % NodeIndexes, GotIt ) + toler

    CALL GetElementNodes( Nodes )

    DO i = 1, n
      Nn = Permutation(Element % NodeIndexes(i))
      IF (Nn==0) CYCLE

      IF (DIM == 2) THEN
        z = Nodes % y( i )
      ELSE IF (DIM == 3) THEN
        z = Nodes % z( i )
      END IF

      ! Geometrical condition
      ! if the node is above the bedrock (plus the tolerance)
      ! then its mask is -1 (floating)
      ! otherwise it is 1 (grounded)

      IF (z > zb(i)) THEN
        VariableValues(Nn) = -1.0_dp
      ELSE
        VariableValues(Nn) = 1.0_dp
      END IF

    END DO
  END DO

  ! attribute zeros to the mask, corresponding to the Grounding Line
  ! loop over each element:
  ! if the sum of the element masks is lower than the element number of nodes minus the number of zeros, then each mask equal to 1 is modified to 0

  DO t = 1, Solver % NumberOfActiveElements

    Element => GetActiveElement(t)
    !IF (ParEnv % myPe .NE. Element % partIndex) CYCLE
    n = GetElementNOFNodes()

    CALL GetElementNodes( Nodes )
    MSum = 0
    ZSum = 0

    DO i = 1, n
      Nn = Permutation(Element % NodeIndexes(i))
      IF (Nn==0) CYCLE
      MSum = MSum + VariableValues(Nn)
      IF (VariableValues(Nn) == 0.0_dp) ZSum = ZSum + 1.0_dp
    END DO

    IF (MSum + ZSum < n) THEN
      DO i = 1, n
        Nn = Permutation(Element % NodeIndexes(i))
        IF (Nn==0) CYCLE

        IF (VariableValues(Nn) == 1.0_dp) THEN
          VariableValues(Nn) = 0.0_dp

          IF (DIM==2) PRINT *, 'Grounding Line, x', Nodes % x( i )
          IF (DIM==3) PRINT *, 'Grounding Line, (x,y)', Nodes % x( i ), Nodes % y( i )

        END IF
      END DO
    END IF
  END DO

  IF ( ParEnv % PEs>1 ) CALL ParallelSumVector( Solver % Matrix, VariableValues, 1 )

  CALL INFO( SolverName , 'Done')

!------------------------------------------------------------------------------
END SUBROUTINE GroundedSolver 
!------------------------------------------------------------------------------


SUBROUTINE GroundedSolverInit( Model,Solver,dt,TransientSimulation )

!DEC$ATTRIBUTES DLLEXPORT :: GroundedSolver
!------------------------------------------------------------------------------
!******************************************************************************
!
! for Grounded Mask initialisation purpose
! same method than above
!
!******************************************************************************
  USE DefUtils

  IMPLICIT NONE
!------------------------------------------------------------------------------
  TYPE(Solver_t) :: Solver
  TYPE(Model_t) :: Model

  REAL(KIND=dp) :: dt
  LOGICAL :: TransientSimulation

!------------------------------------------------------------------------------
! Local variables
!------------------------------------------------------------------------------

  TYPE(Element_t),POINTER :: Element
  TYPE(ValueList_t), POINTER :: Material, SolverParams
  TYPE(Variable_t), POINTER :: PointerToVariable
  TYPE(Nodes_t), SAVE :: Nodes

  LOGICAL :: stat, FirstTime = .TRUE., GotIt, Quadratic

  INTEGER :: i, mn, n, t, Nn, istat, DIM, MSum, ZSum
  INTEGER, POINTER :: Permutation(:)

  REAL(KIND=dp), POINTER :: VariableValues(:)
  REAL(KIND=dp) :: z, toler
  REAL(KIND=dp), ALLOCATABLE :: zb(:)

  CHARACTER(LEN=MAX_NAME_LEN) :: SolverName = 'GroundedSolverInit'

  SAVE DIM, SolverName

  !------------------------------------------------------------------------------

  PointerToVariable => Solver % Variable
  Permutation  => PointerToVariable % Perm
  VariableValues => PointerToVariable % Values

  CALL INFO( SolverName , 'Initializing GroundedMask' )

  !--------------------------------------------------------------
  !Allocate some permanent storage, this is done first time only:
  !--------------------------------------------------------------

  IF (FirstTime) THEN
    FirstTime = .FALSE.
    DIM = CoordinateSystemDimension()
    mn = Solver % Mesh % MaxElementNodes

    ALLOCATE(zb(mn), STAT=istat)

    IF ( istat /= 0 ) THEN
      CALL FATAL( SolverName, 'Memory allocation error.' )
    END IF
    CALL INFO( SolverName, 'Memory allocation done.',Level=1 )

    SolverParams => GetSolverParams()

    toler = GetConstReal(SolverParams, 'TolerInit', GotIt)
    IF (.NOT.GotIt) THEN
      CALL FATAL(SolverName, 'No tolerance given for the Grounded Mask')
    END IF

    ! quadratic elements or not, not used for now
    Quadratic = .FALSE.
    Element => GetActiveElement(1)
    n = GetElementNOFNodes()
    IF ( Element % Type % ElementCode/n == 102 ) Quadratic = .TRUE.

    DO t = 1, Solver % NumberOfActiveElements
      Element => GetActiveElement(t)
      !IF (ParEnv % myPe .NE. Element % partIndex) CYCLE
      n = GetElementNOFNodes()

      Material => GetMaterial( Element )
      zb(1:n) = ListGetReal( Material,'Min Zs Bottom',n , & 
                   Element % NodeIndexes, GotIt ) + toler

      CALL GetElementNodes( Nodes )

      DO i = 1, n
        Nn = Permutation(Element % NodeIndexes(i))
        IF (Nn==0) CYCLE

        IF (DIM == 2) THEN
          z = Nodes % y( i )
        ELSE IF (DIM == 3) THEN
          z = Nodes % z( i )
        END IF

        ! Geometrical condition:
        ! if the node is above the bedrock (plus the tolerance)
        ! then its mask is -1 (floating)
        ! otherwise it is 1 (grounded)
        IF (z > zb(i)) THEN
          VariableValues(Nn) = -1.0_dp
        ELSE
          VariableValues(Nn) = 1.0_dp
        END IF

      END DO

    END DO

    ! attribute zeros to the mask, corresponding to the Grounding Line
    ! loop over each element:
    ! if the sum of the element masks is lower than the element number of nodes minus the number of zeros, then each mask equal to 1 is modified to 0

    DO t = 1, Solver % NumberOfActiveElements

      Element => GetActiveElement(t)
      !IF (ParEnv % myPe .NE. Element % partIndex) CYCLE
      n = GetElementNOFNodes()

      CALL GetElementNodes( Nodes )
      MSum = 0
      ZSum = 0

      DO i = 1, n
        Nn = Permutation(Element % NodeIndexes(i))
        IF (Nn==0) CYCLE

        MSum = MSum + VariableValues(Nn)
        IF (VariableValues(Nn) == 0.0_dp) ZSum = ZSum + 1.0_dp

      END DO

      IF (MSum + ZSum < n) THEN
        DO i = 1, n
          Nn = Permutation(Element % NodeIndexes(i))
          IF (Nn==0) CYCLE

          IF (VariableValues(Nn) == 1.0_dp) THEN
            VariableValues(Nn) = 0.0_dp

            IF (DIM==2) PRINT *, 'Initial Grounding Line, x', Nodes % x( i )
            IF (DIM==3) PRINT *, 'Initial Grounding Line, (x,y)', Nodes % x( i ), Nodes % y( i )

          END IF
        END DO
      END IF

    END DO

    IF ( ParEnv % PEs>1 ) CALL ParallelSumVector( Solver % Matrix, VariableValues, 1 )

    DEALLOCATE(zb)

    CALL INFO( SolverName , 'Done')
    
  END IF

!------------------------------------------------------------------------------
END SUBROUTINE GroundedSolverInit 
!------------------------------------------------------------------------------


